import numpy as np
import torch
import unittest
import rff


class FunctionalTest(unittest.TestCase):

    def test_sample_b(self):
        b = rff.functional.sample_b(1.0, (3, 4))
        self.assertEqual(b.shape, (3, 4))

    def test_gaussian_encoding(self):
        v = rff.dataloader.rectangular_coordinates((2, 2))
        b = torch.eye(2)
        gamma_v = rff.functional.gaussian_encoding(v, b)
        self.assertEqual(gamma_v.shape, (2, 2, 4))
        xc = np.cos(2 * np.pi * v)
        yc = np.sin(2 * np.pi * v)
        gamma_v_expected = np.concatenate((xc, yc), axis=-1)
        np.testing.assert_almost_equal(gamma_v, gamma_v_expected, decimal=5)

    def test_basic_encoding(self):
        v = rff.dataloader.rectangular_coordinates((2, 2))
        gamma_v = rff.functional.basic_encoding(v)
        self.assertEqual(gamma_v.shape, (2, 2, 4))
        xc = np.cos(2 * np.pi * v)
        yc = np.sin(2 * np.pi * v)
        gamma_v_expected = np.concatenate((xc, yc), axis=-1)
        np.testing.assert_almost_equal(gamma_v, gamma_v_expected, decimal=5)

    def test_positional_encoding(self):
        v = rff.dataloader.rectangular_coordinates((2, 2))
        gamma_v = rff.functional.positional_encoding(v, sigma=1.0, m=1)
        self.assertEqual(gamma_v.shape, (2, 2, 4))
        xc = np.cos(2 * np.pi * v)
        yc = np.sin(2 * np.pi * v)
        gamma_v_expected = np.concatenate(
            (xc[..., np.newaxis], yc[..., np.newaxis]), axis=-1)
        gamma_v_expected = np.reshape(gamma_v_expected, (2, 2, 4))
        np.testing.assert_almost_equal(gamma_v, gamma_v_expected, decimal=5)



"""
Test layers, the point to to make sure they can be created and that the output matches
the corresponding functional calls. Other testing will be performed in the functional
test.
"""

import numpy as np
import rff
import pytest
import torch


def check_cuda(device):
    if device == 'cuda' and not torch.cuda.is_available():
        pytest.skip('Cuda is not available')


@pytest.mark.parametrize('device', ['cpu', 'cuda'])
def test_gaussian_encoding(device):
    check_cuda(device)
    b = rff.functional.sample_b(1.0, (256, 2)).to(device)
    layer = rff.layers.GaussianEncoding(b=b).to(device)
    v = rff.dataloader.rectangular_coordinates((256, 256)).to(device)
    gamma_v = layer(v)
    gamma_v_expected = rff.functional.gaussian_encoding(v, b)
    np.testing.assert_almost_equal(
        gamma_v.cpu().numpy(),
        gamma_v_expected.cpu().numpy(),
        decimal=5)


@pytest.mark.parametrize('device', ['cpu', 'cuda'])
def test_basic_encoding(device):
    check_cuda(device)
    layer = rff.layers.BasicEncoding().to(device)
    v = rff.dataloader.rectangular_coordinates((256, 256)).to(device)
    gamma_v = layer(v)
    gamma_v_expected = rff.functional.basic_encoding(v)
    np.testing.assert_almost_equal(
        gamma_v.cpu().numpy(),
        gamma_v_expected.cpu().numpy(),
        decimal=5)


@pytest.mark.parametrize('device', ['cpu', 'cuda'])
def test_positional_encoding(device):
    check_cuda(device)
    layer = rff.layers.PositionalEncoding(sigma=1.0, m=10).to(device)
    v = rff.dataloader.rectangular_coordinates((256, 256)).to(device)
    gamma_v = layer(v)
    gamma_v_expected = rff.functional.positional_encoding(
        v, sigma=1.0, m=10)
    np.testing.assert_almost_equal(
        gamma_v.cpu().numpy(),
        gamma_v_expected.cpu().numpy(),
        decimal=5)



import torch
import torchvision

from torch import Tensor
from torch.utils.data.dataset import TensorDataset


def rectangular_coordinates(size: tuple) -> Tensor:
    r"""Creates a tensor of equally spaced coordinates for use with an image or volume

    Args:
        size (tuple): shape of the image or volume

    Returns:
        Tensor: tensor of shape :math:`(*\text{size}, \text{len(size)})`
    """
    def linspace_func(nx): return torch.linspace(0.0, 1.0, nx)
    linspaces = map(linspace_func, size)
    coordinates = torch.meshgrid(*linspaces)
    return torch.stack(coordinates, dim=-1)


def to_dataset(path: str) -> TensorDataset:
    image = torchvision.io.read_image(path).float()
    _, H, W = image.shape
    coords = rectangular_coordinates((H, W))
    image = image.permute((1, 2, 0))
    image /= 255.0
    coords = coords.flatten(0, -2)
    image = image.flatten(0, -2)
    return TensorDataset(coords, image)



import numpy as np
import unittest
import rff


class DataloaderTest(unittest.TestCase):

    def test_rectangular_coordinates_shape(self):
        size = (3, 4, 5)
        coords = rff.dataloader.rectangular_coordinates(size)
        self.assertEqual(coords.shape, (*size, len(size)))

    def test_rectangular_coordinates_value(self):
        size = (2, 2)
        coords = rff.dataloader.rectangular_coordinates(size)
        expected_coords = np.array([[[0, 0], [0, 1]], [[1, 0], [1, 1]]])
        np.testing.assert_equal(coords, expected_coords)

    def test_to_dataset(self):
        _ = rff.dataloader.to_dataset('images/cat.jpg')



import torch.nn as nn

from typing import Optional
from torch import Tensor
import rff


class GaussianEncoding(nn.Module):
    """Layer for mapping coordinates using random Fourier features"""

    def __init__(self, sigma: Optional[float] = None,
                 input_size: Optional[float] = None,
                 encoded_size: Optional[float] = None,
                 b: Optional[Tensor] = None):
        r"""
        Args:
            sigma (Optional[float]): standard deviation
            input_size (Optional[float]): the number of input dimensions
            encoded_size (Optional[float]): the number of dimensions the `b` matrix maps to
            b (Optional[Tensor], optional): Optionally specify a :attr:`b` matrix already sampled
        Raises:
            ValueError:
                If :attr:`b` is provided and one of :attr:`sigma`, :attr:`input_size`,
                or :attr:`encoded_size` is provided. If :attr:`b` is not provided and one of
                :attr:`sigma`, :attr:`input_size`, or :attr:`encoded_size` is not provided.
        """
        super().__init__()
        if b is None:
            if sigma is None or input_size is None or encoded_size is None:
                raise ValueError(
                    'Arguments "sigma," "input_size," and "encoded_size" are required.')

            b = rff.functional.sample_b(sigma, (encoded_size, input_size))
        elif sigma is not None or input_size is not None or encoded_size is not None:
            raise ValueError('Only specify the "b" argument when using it.')
        self.b = nn.parameter.Parameter(b, requires_grad=False)

    def forward(self, v: Tensor) -> Tensor:
        r"""Computes :math:`\gamma(\mathbf{v}) = (\cos{2 \pi \mathbf{B} \mathbf{v}} , \sin{2 \pi \mathbf{B} \mathbf{v}})`

        Args:
            v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`

        Returns:
            Tensor: Tensor mapping using random fourier features of shape :math:`(N, *, 2 \cdot \text{encoded_size})`
        """
        return rff.functional.gaussian_encoding(v, self.b)


class BasicEncoding(nn.Module):
    """Layer for mapping coordinates using the basic encoding"""

    def forward(self, v: Tensor) -> Tensor:
        r"""Computes :math:`\gamma(\mathbf{v}) = (\cos{2 \pi \mathbf{v}} , \sin{2 \pi \mathbf{v}})`

        Args:
            v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`

        Returns:
            Tensor: mapped tensor of shape :math:`(N, *, 2 \cdot \text{input_size})`
        """
        return rff.functional.basic_encoding(v)


class PositionalEncoding(nn.Module):
    """Layer for mapping coordinates using the positional encoding"""

    def __init__(self, sigma: float, m: int):
        r"""
        Args:
            sigma (float): frequency constant
            m (int): number of frequencies to map to
        """
        super().__init__()
        self.sigma = sigma
        self.m = m

    def forward(self, v: Tensor) -> Tensor:
        r"""Computes :math:`\gamma(\mathbf{v}) = (\dots, \cos{2 \pi \sigma^{(j/m)} \mathbf{v}} , \sin{2 \pi \sigma^{(j/m)} \mathbf{v}}, \dots)`

        Args:
            v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`

        Returns:
            Tensor: mapped tensor of shape :math:`(N, *, 2 \cdot m \cdot \text{input_size})`
        """
        return rff.functional.positional_encoding(v, self.sigma, self.m)



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='random-fourier-features-pytorch',
      version='0.2',
      description='Random Fourier Features for PyTorch',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/jmclong/random-fourier-features-pytorch',
      author='Joshua Long',
      author_email='joshualong@live.com',
      license='MIT',
      packages=['rff'],
      install_requires=["numpy>=1.17",
                        "torch>=1.7",
                        "torchvision>=0.9"])



import numpy as np
import torch

from torch import Tensor


def sample_b(sigma: float, size: tuple) -> Tensor:
    r"""Matrix of size :attr:`size` sampled from from :math:`\mathcal{N}(0, \sigma^2)`

    Args:
        sigma (float): standard deviation
        size (tuple) size of the matrix sampled

    See :class:`~rff.layers.GaussianEncoding` for more details
    """
    return torch.randn(size) * sigma


@torch.jit.script
def gaussian_encoding(
        v: Tensor,
        b: Tensor) -> Tensor:
    r"""Computes :math:`\gamma(\mathbf{v}) = (\cos{2 \pi \mathbf{B} \mathbf{v}} , \sin{2 \pi \mathbf{B} \mathbf{v}})`

    Args:
        v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`
        b (Tensor): projection matrix of shape :math:`(\text{encoded_layer_size}, \text{input_size})`

    Returns:
        Tensor: mapped tensor of shape :math:`(N, *, 2 \cdot \text{encoded_layer_size})`

    See :class:`~rff.layers.GaussianEncoding` for more details.
    """
    vp = 2 * np.pi * v @ b.T
    return torch.cat((torch.cos(vp), torch.sin(vp)), dim=-1)


@torch.jit.script
def basic_encoding(
        v: Tensor) -> Tensor:
    r"""Computes :math:`\gamma(\mathbf{v}) = (\cos{2 \pi \mathbf{v}} , \sin{2 \pi \mathbf{v}})`

    Args:
        v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`

    Returns:
        Tensor: mapped tensor of shape :math:`(N, *, 2 \cdot \text{input_size})`

    See :class:`~rff.layers.BasicEncoding` for more details.
    """
    vp = 2 * np.pi * v
    return torch.cat((torch.cos(vp), torch.sin(vp)), dim=-1)


@torch.jit.script
def positional_encoding(
        v: Tensor,
        sigma: float,
        m: int) -> Tensor:
    r"""`\gamma(\mathbf{v}) = (\dots, \cos{2 \pi \sigma^{(j/m)} \mathbf{v}} , \sin{2 \pi \sigma^{(j/m)} \mathbf{v}}, \dots)`
        where :math:`j \in \{0, \dots, m-1\}`

    Args:
        v (Tensor): input tensor of shape :math:`(N, *, \text{input_size})`
        sigma (float): constant chosen based upon the domain of :attr:`v`
        m (int): [description]

    Returns:
        Tensor: mapped tensor of shape :math:`(N, *, 2 \cdot m \cdot \text{input_size})`

    See :class:`~rff.layers.PositionalEncoding` for more details.
    """
    j = torch.arange(m, device=v.device)
    coeffs = 2 * np.pi * sigma ** (j / m)
    vp = coeffs * torch.unsqueeze(v, -1)
    vp_cat = torch.cat((torch.cos(vp), torch.sin(vp)), dim=-1)
    return vp_cat.flatten(-2, -1)
